import logging
import os
import pathlib
from logging.handlers import TimedRotatingFileHandler

import yaml

from typing import Dict, Optional

from metricflow.cli.constants import (
    CONFIG_DWH_DB,
    CONFIG_DWH_DIALECT,
    CONFIG_DWH_HOST,
    CONFIG_DWH_PASSWORD,
    CONFIG_DWH_PORT,
    CONFIG_DWH_SCHEMA,
    CONFIG_DWH_USER,
    CONFIG_DWH_WAREHOUSE,
    CONFIG_MODEL_PATH,
    CONFIG_PATH_KEY,
)
from metricflow.cli.exceptions import SqlClientException, MetricFlowInitException, ModelCreationException
from metricflow.cli.time_source import ServerTimeSource
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.model.parsing.dir_to_model import parse_directory_of_yaml_files_to_model
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient, SupportedSqlEngine
from metricflow.sql_clients.common_client import not_empty
from metricflow.sql_clients.big_query import BigQuerySqlClient
from metricflow.sql_clients.redshift import RedshiftSqlClient
from metricflow.sql_clients.snowflake import SnowflakeSqlClient

logger = logging.getLogger(__name__)


class CLIContext:
    """Context for MetricFlow CLI."""

    def __init__(self) -> None:  # noqa: D
        self.verbose = False
        self._mf: Optional[MetricFlowEngine] = None
        self._sql_client: Optional[SqlClient] = None
        self._semantic_model: Optional[SemanticModel] = None
        self._mf_system_schema: Optional[str] = None
        self.config = ConfigHandler()
        self._configure_logging()

    def _configure_logging(self) -> None:  # noqa: D
        log_format = "%(asctime)s %(levelname)s %(filename)s:%(lineno)d [%(threadName)s] - %(message)s"
        logging.basicConfig(level=logging.INFO, format=log_format)

        log_file_handler = TimedRotatingFileHandler(
            filename=self.config.log_file_path,
            # Rotate every day to a new file, keep 7 days worth.
            when="D",
            interval=1,
            backupCount=7,
        )

        formatter = logging.Formatter(fmt=log_format)
        log_file_handler.setFormatter(formatter)
        log_file_handler.setLevel(logging.INFO)

        root_logger = logging.getLogger()
        # StreamHandler to the console would have been setup by logging.basicConfig
        for handler in root_logger.handlers:
            handler.setLevel(logging.CRITICAL)
        root_logger.addHandler(log_file_handler)

    @property
    def mf_system_schema(self) -> str:  # noqa: D
        if self._mf_system_schema is None:
            self._mf_system_schema = self.config.get_config_value(CONFIG_DWH_SCHEMA)
        assert self._mf_system_schema
        return self._mf_system_schema

    def __initialize_sql_client(self) -> None:
        """Initializes the SqlClient given the credentials."""
        try:
            dialect = self.config.get_config_value(CONFIG_DWH_DIALECT).upper()
            url = f"file://{self.config.file_path}"
            if dialect == SupportedSqlEngine.BIGQUERY.name:
                path_to_creds = not_empty(self.config.get_config_value(CONFIG_DWH_PASSWORD), CONFIG_DWH_PASSWORD, url)
                if not pathlib.Path(path_to_creds).exists:
                    raise ValueError(f"`{path_to_creds}` does not contain the BigQuery credential file.")
                with open(path_to_creds, "r") as cred_file:
                    creds = cred_file.readline()
                self._sql_client = BigQuerySqlClient(password=creds)
            elif dialect == SupportedSqlEngine.SNOWFLAKE.name:
                host = not_empty(self.config.get_config_value(CONFIG_DWH_HOST), CONFIG_DWH_HOST, url)
                user = not_empty(self.config.get_config_value(CONFIG_DWH_USER), CONFIG_DWH_USER, url)
                password = self.config.get_config_value(CONFIG_DWH_PASSWORD)
                database = not_empty(self.config.get_config_value(CONFIG_DWH_DB), CONFIG_DWH_DB, url)
                warehouse = not_empty(self.config.get_config_value(CONFIG_DWH_WAREHOUSE), CONFIG_DWH_WAREHOUSE, url)
                self._sql_client = SnowflakeSqlClient(
                    host=host,
                    username=user,
                    password=password,
                    database=database,
                    url_query_params={"warehouse": warehouse},
                )
            elif dialect == SupportedSqlEngine.REDSHIFT.name:
                host = not_empty(self.config.get_config_value(CONFIG_DWH_HOST), CONFIG_DWH_HOST, url)
                port = int(self.config.get_config_value(CONFIG_DWH_PORT))
                user = not_empty(self.config.get_config_value(CONFIG_DWH_USER), CONFIG_DWH_USER, url)
                password = self.config.get_config_value(CONFIG_DWH_PASSWORD)
                database = not_empty(self.config.get_config_value(CONFIG_DWH_DB), CONFIG_DWH_DB, url)
                self._sql_client = RedshiftSqlClient(
                    host=host,
                    port=port,
                    username=user,
                    password=password,
                    database=database,
                )
            else:
                raise ValueError(
                    f"Invalid dialect `{dialect}`, must be one of `bigquery`, `snowflake`, `redshift` in {url}"
                )
        except Exception as e:
            raise SqlClientException from e

    @property
    def sql_client(self) -> SqlClient:  # noqa: D
        if self._sql_client is None:
            # Initialize the SqlClient if not set
            self.__initialize_sql_client()
        assert self._sql_client is not None
        return self._sql_client

    def run_health_checks(self) -> Dict[str, Dict[str, str]]:
        """Execute the DB health checks."""
        try:
            return self.sql_client.health_checks(self.mf_system_schema)
        except Exception:
            raise SqlClientException

    def _initialize_metricflow_engine(self) -> None:
        """Initialize the MetricFlowEngine."""
        try:
            self._mf = MetricFlowEngine(
                semantic_model=self.semantic_model,
                sql_client=self.sql_client,
                column_association_resolver=DefaultColumnAssociationResolver(self.semantic_model),
                time_source=ServerTimeSource(),
                time_spine_source=TimeSpineSource(self.sql_client, schema_name=self.mf_system_schema),
                system_schema=self.mf_system_schema,
            )
        except Exception as e:
            raise MetricFlowInitException from e

    @property
    def mf(self) -> MetricFlowEngine:  # noqa: D
        if self._mf is None:
            self._initialize_metricflow_engine()
        assert self._mf is not None
        return self._mf

    def _build_semantic_model(self) -> None:
        """Get the path to the models and create a corresponding SemanticModel."""
        path_to_models = self.config.get_config_value(CONFIG_MODEL_PATH)
        try:
            model = parse_directory_of_yaml_files_to_model(path_to_models).model
            assert model is not None
        except Exception as e:
            raise ModelCreationException from e
        self._semantic_model = SemanticModel(model)

    @property
    def semantic_model(self) -> SemanticModel:  # noqa: D
        if self._semantic_model is None:
            self._build_semantic_model()
        assert self._semantic_model is not None
        return self._semantic_model


class ConfigHandler:
    """Class to handle all config file retrieval/insertion actions."""

    def __init__(self) -> None:  # noqa: D
        # Create config directory if not exist
        if not os.path.exists(self.dir_path):
            dir_path = pathlib.Path(self.dir_path)
            dir_path.mkdir(parents=True)

    @property
    def dir_path(self) -> str:
        """Retrieve MetricFlow config directory from $MF_CONFIG_DIR, default config dir is ~/.metricflow"""
        config_dir_env = os.getenv(CONFIG_PATH_KEY)
        return config_dir_env if config_dir_env else f"{str(pathlib.Path.home())}/.metricflow"

    @property
    def file_path(self) -> str:
        """Config file can be found at <config_dir>/config.yml"""
        return os.path.join(self.dir_path, "config.yml")

    def get_config_value(self, key: str) -> str:
        """Attempt to get a corresponding value from the config file. Throw an error if not exists or None."""
        config_file = open(self.file_path, "r")

        config = yaml.load(config_file, Loader=yaml.Loader)
        if key not in config:
            raise KeyError(f"Key '{key}' is missing in the configuration file '{self.file_path}'")

        value = config[key]
        if value is None:
            raise ValueError(f"Value for key '{key}' cannot be none in the configuration file '{self.file_path}")
        return value

    @property
    def log_file_path(self) -> str:
        """Returns the name of the log file where all logging messages are stored."""
        log_dir_elements = [self.dir_path, "logs"]
        log_dir = os.path.join(*log_dir_elements)
        os.makedirs(log_dir, exist_ok=True)
        return os.path.join(*(log_dir_elements + ["metricflow_cli.log"]))



"""File containing constants used by CLI."""

# Configuration Constants
CONFIG_PATH_KEY = "MF_CONFIG_DIR"
CONFIG_DWH_DIALECT = "dwh_dialect"
CONFIG_DWH_HOST = "dwh_host"
CONFIG_DWH_PORT = "dwh_port"
CONFIG_DWH_USER = "dwh_user"
CONFIG_DWH_PASSWORD = "dwh_password"
CONFIG_DWH_DB = "dwh_database"
CONFIG_DWH_SCHEMA = "dwh_schema"
CONFIG_DWH_WAREHOUSE = "dwh_warehouse"
CONFIG_MODEL_PATH = "model_path"

# Display Setting Constants
MAX_LIST_OBJECT_ELEMENTS = 5
DEFAULT_RESULT_DECIMAL_PLACES = 2



from datetime import datetime
from unittest.mock import patch, MagicMock

from metricflow.cli.main import (
    drop_materialization,
    get_dimension_values,
    list_dimensions,
    list_materializations,
    list_metrics,
    materialize,
    query,
    validate_configs,
)
from metricflow.model.model_validator import ModelValidator
from metricflow.model.validations.validator_helpers import (
    ValidationError,
    ValidationFatal,
    ValidationFutureError,
    ValidationWarning,
)
from metricflow.test.fixtures.cli_fixtures import MetricFlowCliRunner


def test_query(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(query, args=["--metrics", "metric1", "--dimensions", "ds"])

    assert "metric1" in resp.output
    assert resp.exit_code == 0


def test_list_dimensions(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_dimensions, args=["--metric-names", "metric1"])

    assert "dim1" in resp.output
    assert resp.exit_code == 0


def test_list_metrics(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_metrics)

    assert "metric1: dim1, dim2, dim3" in resp.output
    assert resp.exit_code == 0


def test_get_dimension_values(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(get_dimension_values, args=["--metric-name", "metric1", "--dimension-name", "ds"])

    assert "dim_val1" in resp.output
    assert resp.exit_code == 0


def test_list_materializations(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(list_materializations)

    assert "materialization: details related to materialization" in resp.output
    assert resp.exit_code == 0


def test_materialize(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(materialize, args=["--materialization-name", "test", "--start-time", "2020-01-02"])

    assert "test.table" in resp.output
    assert resp.exit_code == 0


def test_drop_materialization(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    resp = cli_runner.run(drop_materialization, args=["--materialization-name", "test"])

    assert resp.exit_code == 0


def test_validate_configs(cli_runner: MetricFlowCliRunner) -> None:  # noqa: D
    # Mock validate_model function
    mocked_build_result = MagicMock(
        issues=(
            ValidationWarning(None, "warning"),  # type: ignore
            ValidationFutureError(None, "future_error", datetime.now()),  # type: ignore
            ValidationError(None, "error"),  # type: ignore
            ValidationFatal(None, "fatal"),  # type: ignore
        )
    )
    with patch.object(ModelValidator, "validate_model", return_value=mocked_build_result):
        resp = cli_runner.run(validate_configs)

    assert "future_error" in resp.output
    assert resp.exit_code == 0



import datetime as dt
import logging
import pathlib
import traceback
from functools import wraps
from typing import Any, Callable, List, Optional

import click
from dateutil.parser import parse

from metricflow.cli.cli_context import CLIContext
from metricflow.cli.config_builder import ConfigKey
from metricflow.cli.constants import (
    CONFIG_DWH_DB,
    CONFIG_DWH_DIALECT,
    CONFIG_DWH_HOST,
    CONFIG_DWH_PASSWORD,
    CONFIG_DWH_PORT,
    CONFIG_DWH_SCHEMA,
    CONFIG_DWH_USER,
    CONFIG_DWH_WAREHOUSE,
    CONFIG_MODEL_PATH,
)
from metricflow.model.validations.validator_helpers import ValidationIssueLevel

logger = logging.getLogger(__name__)

# MetricFlow config keys
MF_CONFIG_KEYS = [
    ConfigKey(
        key=CONFIG_MODEL_PATH,
        value=f"{pathlib.Path.home()}/.metricflow/sample_models",
        comment="Path to directory containing defined models (Leave until after DWH setup)",
    ),
    ConfigKey(key=CONFIG_DWH_SCHEMA),
]
# BigQuery config keys
MF_BIGQUERY_KEYS = [
    ConfigKey(key=CONFIG_DWH_PASSWORD, comment="Provide the path to the BigQuery credential file"),
    ConfigKey(key=CONFIG_DWH_DIALECT, value="bigquery", comment="Dialect (one of BigQuery, Snowflake, Redshift)"),
]
# Redshift config keys
MF_REDSHIFT_KEYS = [
    ConfigKey(key=CONFIG_DWH_DB),
    ConfigKey(key=CONFIG_DWH_PASSWORD, comment="Password associated with the provided user"),
    ConfigKey(key=CONFIG_DWH_USER, comment="Username for the data warehouse"),
    ConfigKey(key=CONFIG_DWH_PORT),
    ConfigKey(key=CONFIG_DWH_HOST, comment="Host name"),
    ConfigKey(key=CONFIG_DWH_DIALECT, value="redshift", comment="Dialect (one of BigQuery, Snowflake, Redshift)"),
]
# Redshift config keys
MF_SNOWFLAKE_KEYS = [
    ConfigKey(key=CONFIG_DWH_WAREHOUSE, comment="Provide the warehouse to use"),
    ConfigKey(key=CONFIG_DWH_DB),
    ConfigKey(key=CONFIG_DWH_PASSWORD, comment="Password associated with the provided user"),
    ConfigKey(key=CONFIG_DWH_USER, comment="Username for the data warehouse"),
    ConfigKey(key=CONFIG_DWH_HOST, comment="Snowflake account name"),
    ConfigKey(key=CONFIG_DWH_DIALECT, value="snowflake", comment="Dialect (one of BigQuery, Snowflake, Redshift)"),
]


# Click Options
def query_options(function: Callable) -> Callable:
    """Common options for a query"""
    function = click.option(
        "--order",
        required=False,
        multiple=True,
        help='Metrics or dimensions to order by ("-" in front of a column means descending). For example: --order -ds',
    )(function)

    function = click.option(
        "--limit",
        required=False,
        type=str,
        help="Limit the number of rows out using an int or leave blank for no limit. For example: --limit 100",
        callback=lambda ctx, param, value: validate_limit(value),
    )(function)

    function = click.option(
        "--where",
        required=False,
        type=str,
        default=None,
        help='SQL-like where statement provided as a string. For example: --where "revenue > 100"',
    )(function)

    function = start_end_time_options(function)
    return function


def start_end_time_options(function: Callable) -> Callable:
    """Options for start_time and end_time."""
    function = click.option(
        "--start-time",
        required=False,
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the start time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)

    function = click.option(
        "--end-time",
        required=False,
        type=str,
        default=None,
        help="Optional iso8601 timestamp to constraint the end time of the data (inclusive)",
        callback=lambda ctx, param, value: convert_to_datetime(value),
    )(function)
    return function


def separated_by_comma_option(option_name: str, help_msg: str = "") -> Callable:
    """Parse input containing a string separated by commma to a List."""

    def wraps(function: Callable) -> Callable:
        function = click.option(
            option_name,
            required=True,
            help=help_msg,
            callback=lambda ctx, param, value: parse_comma_separated_inputs(value),
        )(function)
        return function

    return wraps


# Parsers/Validators
def convert_to_datetime(datetime_str: Optional[str]) -> Optional[dt.datetime]:
    """Callback to convert string to datetime given as an iso8601 timestamp."""
    if datetime_str is None:
        return None

    if not valid_datetime(datetime_str):
        raise click.BadParameter("must be valid iso8601 timestamp")

    return dt.datetime.fromisoformat(datetime_str)


def valid_datetime(dt_str: str) -> bool:
    """Returns true if string is valid iso8601 timestamp, false otherwise"""
    try:
        parse(dt_str)
    except Exception:
        return False
    return True


def parse_comma_separated_inputs(value: str) -> List[str]:  # noqa: D
    # If comma exist, explode this into a list and return
    if "," in value:
        return [i.strip() for i in value.split(",")]

    # Return a list of the single value
    return [value]


def validate_limit(limit: Optional[str]) -> Optional[int]:
    """Validates and transform limit input."""
    if limit and not limit.isnumeric():
        raise click.BadParameter("limit must be an int. For no limit, do not pass this argument")
    return int(limit) if limit else None


# Misc
def exception_handler(func: Callable[..., Any]) -> Callable[..., Any]:  # type: ignore[misc]
    """Decorator to handle exceptions."""

    @wraps(func)
    def wrapper(*args, **kwargs):  # type: ignore
        try:
            func(*args, **kwargs)
        except Exception as e:
            # This will log to the file handlers registered in the root.
            logging.exception("Got an exception in the exception handler.")
            # Checks if CLIContext has verbose flag set

            if isinstance(args[0], CLIContext):
                cli_context: CLIContext = args[0]
                click.echo(f"\nERROR: {str(e)} - Log file: {cli_context.config.log_file_path}")
            else:
                if not isinstance(args[0], CLIContext):
                    logger.error(
                        f"Missing {CLIContext.__name__} as the first argument to the function "
                        f"{getattr(func, '__name__', repr(func))}"
                    )
                click.echo(f"\nERROR: {str(e)}")
            if args and hasattr(args[0], "verbose") and args[0].verbose is True:
                click.echo(traceback.format_exc())

            exit(1)

    return wrapper


def build_validation_header_msg(level: ValidationIssueLevel) -> str:
    """Builds the header message with colour."""
    colour_map = {
        ValidationIssueLevel.WARNING: "cyan",
        ValidationIssueLevel.ERROR: "bright_red",
        ValidationIssueLevel.FATAL: "bright_red",
        ValidationIssueLevel.FUTURE_ERROR: "bright_yellow",
    }

    return click.style(level.name, bold=True, fg=colour_map[level])
